import os
import json
import datetime

from google.cloud import datastore

from cirrocumulus.abstract_db import AbstractDB
from cirrocumulus.envir import CIRRO_EMAIL, CIRRO_JOB_RESULTS, SERVER_CAPABILITY_JOBS
from cirrocumulus.invalid_usage import InvalidUsage
from cirrocumulus.job_api import save_job_result_to_file
from cirrocumulus.util import get_email_domain


DATASET = "Dataset"
CAT_NAME = "Cat_Name"
DATASET_FILTER = "Dataset_Filter"
DATASET_VIEW = "D_View"
FEATURE_SET = "F_Set"
USER = "User"
JOB = "Job"


def get_datasets(results, email, query, unique_ids):
    for result in query.fetch():
        if result.id not in unique_ids:
            unique_ids.add(result.id)
            results.append(
                {
                    "id": result.id,
                    "name": result["name"],
                    "title": result.get("title"),
                    "owner": email in result["owners"],
                    "species": result.get("species"),
                    "url": result["url"],
                }
            )


class CloudFireStore(AbstractDB):
    def __init__(self):
        super().__init__()
        self.datastore_client = datastore.Client()
        os.environ[CIRRO_EMAIL] = self.datastore_client.project + "@appspot.gserviceaccount.com"

    def capabilities(self):
        c = super().capabilities()
        c[SERVER_CAPABILITY_JOBS] = False
        return c

    def __get_key_and_dataset(self, email, dataset_id, ensure_owner=False):
        client = self.datastore_client
        key = client.key(DATASET, int(dataset_id))
        dataset = client.get(key)
        if dataset is None:
            raise InvalidUsage("Please provide a valid id", 400)
        readers = dataset.get("readers")

        domain = get_email_domain(email)
        if email not in readers and domain not in readers:
            raise InvalidUsage("Not authorized", 403)
        if ensure_owner and email not in dataset["owners"]:
            raise InvalidUsage("Not authorized", 403)
        return key, dataset

    def __get_entity_list(self, email, dataset_id, kind, keys):
        client = self.datastore_client
        self.__get_key_and_dataset(email, dataset_id)
        query = client.query(kind=kind)
        query.add_filter("dataset_id", "=", int(dataset_id))
        results = []
        for result in query.fetch():
            d = dict(id=result.id)
            for key in keys:
                d[key] = result.get(key)
            results.append(result)
        return results

    def __delete_entity(self, email, kind, entity_id):
        client = self.datastore_client
        key = client.key(kind, int(entity_id))
        e = client.get(key)
        self.__get_key_and_dataset(email, e["dataset_id"])
        client.delete(key)

    def __get_entity(self, email, kind, entity_id):
        client = self.datastore_client
        e = client.get(client.key(kind, int(entity_id)))
        self.__get_key_and_dataset(email, e["dataset_id"])
        return e

    def __upsert_entity(self, email, dataset_id, entity_id, kind, entity_update):
        dataset_id = int(dataset_id)
        client = self.datastore_client
        self.__get_key_and_dataset(email, dataset_id)
        if entity_id is None:
            e = datastore.Entity(client.key(kind), exclude_from_indexes=["value"])
        else:
            entity_id = int(entity_id)
            key = client.key(kind, entity_id)
            e = client.get(key)

        if email is not None:
            entity_update["email"] = email
        if dataset_id is not None:
            entity_update["dataset_id"] = dataset_id

        e.update(entity_update)
        client.put(e)
        return e.id

    def category_names(self, email, dataset_id):
        dataset_id = int(dataset_id)
        self.__get_key_and_dataset(email, dataset_id, False)
        client = self.datastore_client
        query = client.query(kind=CAT_NAME)
        query.add_filter("dataset_id", "=", dataset_id)
        results = {}
        for result in query.fetch():
            category = result["category"]
            cat_results = results.get(category)
            if cat_results is None:
                cat_results = {}
                results[category] = cat_results
            cat_results[result["original"]] = {
                "newValue": result.get("new"),
                "negativeMarkers": result.get("negativeMarkers"),
                "positiveMarkers": result.get("positiveMarkers"),
                "color": result.get("color"),
            }

        return results

    def upsert_category_name(self, email, dataset_id, category, original_value, update):
        dataset_id = int(dataset_id)
        self.__get_key_and_dataset(email, dataset_id, False)
        client = self.datastore_client
        key = client.key(
            CAT_NAME, str(dataset_id) + "-" + str(category) + "-" + str(original_value)
        )
        entity = datastore.Entity(key=key)
        update["category"] = category
        update["dataset_id"] = dataset_id
        update["original"] = original_value
        if "newValue" in update:  # backwards compatibility
            update["new"] = update.pop("newValue")
        entity.update(update)
        client.put(entity)
        return entity.id

    def user(self, email):
        client = self.datastore_client
        key = client.key(USER, email)
        user = client.get(key)
        if user is None:
            user = datastore.Entity(client.key(USER, email))
        user.update({"last_login": datetime.datetime.utcnow()})
        client.put(user)
        return user

    def datasets(self, email):
        client = self.datastore_client
        query = client.query(kind=DATASET)
        query.add_filter("readers", "=", email)
        results = []
        unique_ids = set()
        get_datasets(results, email, query, unique_ids)
        domain = get_email_domain(email)
        if domain is not None:
            query = client.query(kind=DATASET)
            query.add_filter("readers", "=", domain)
            # TODO make this more efficient
            get_datasets(results, email, query, unique_ids)

            query = client.query(kind=DATASET)
            query.add_filter("readers", "=", "allUsers")
            get_datasets(results, email, query, unique_ids)
        return results

    def delete_dataset(self, email, dataset_id):
        client = self.datastore_client
        key, dataset = self.__get_key_and_dataset(email, dataset_id, True)
        client.delete(key)

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        _, dataset = self.__get_key_and_dataset(email, dataset_id, ensure_owner)
        dataset["id"] = dataset.id
        return dataset

    def is_importer(self, email):
        # TODO check if user can modify dataset
        client = self.datastore_client
        user = client.get(client.key(USER, email))
        if "importer" not in user:
            raise False
        return user["importer"]

    def upsert_dataset(self, email, readers, dataset):
        client = self.datastore_client

        if readers is not None:
            readers = set(readers)
            if email in readers:
                readers.remove(email)
            readers.add(email)
            dataset["readers"] = list(readers)
        dataset_id = dataset.get("id")
        if dataset_id is not None:  # only owner can update
            key, dataset_entity = self.__get_key_and_dataset(email, dataset_id, True)
        else:  # new dataset
            dataset_entity = datastore.Entity(client.key(DATASET), exclude_from_indexes=["url"])
            user = client.get(client.key(USER, email))
            if "importer" not in user or not user["importer"]:
                raise InvalidUsage("Not authorized", 403)
            dataset["owners"] = [email]

        dataset_entity.update(dataset)
        client.put(dataset_entity)
        dataset_id = dataset_entity.id
        return dataset_id

    # feature sets
    def get_feature_sets(self, email, dataset_id):
        return self.__get_entity_list(
            email=email,
            dataset_id=dataset_id,
            kind=FEATURE_SET,
            keys=["category", "name", "features"],
        )

    def delete_feature_set(self, email, dataset_id, set_id):
        return self.__delete_entity(email=email, kind=FEATURE_SET, entity_id=set_id)

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        entity_update = {}
        if name is not None:
            entity_update["name"] = name
        if category is not None:
            entity_update["category"] = category
        if features is not None:
            entity_update["features"] = features
        return self.__upsert_entity(
            email=email,
            dataset_id=dataset_id,
            entity_id=set_id,
            kind=FEATURE_SET,
            entity_update=entity_update,
        )

    # views
    def dataset_views(self, email, dataset_id):
        return self.__get_entity_list(
            email=email,
            dataset_id=dataset_id,
            kind=DATASET_VIEW,
            keys=["name", "notes", "email", "last_updated"],
        )

    def delete_dataset_view(self, email, view_id):
        return self.__delete_entity(email=email, kind=DATASET_VIEW, entity_id=view_id)

    def get_dataset_view(self, email, view_id):
        return self.__get_entity(email=email, entity_id=view_id, kind=DATASET_VIEW)

    def upsert_dataset_view(self, email, dataset_id, view):
        view["last_updated"] = datetime.datetime.utcnow()
        if "value" in view:
            view["value"] = json.dumps(view["value"])
        if email is not None:
            view["email"] = email
        if dataset_id is not None:
            view["dataset_id"] = dataset_id
        view_id = view.pop("id") if "id" in view else None
        view_id = self.__upsert_entity(
            email=email,
            dataset_id=dataset_id,
            entity_id=view_id,
            kind=DATASET_VIEW,
            entity_update=view,
        )
        return dict(id=view_id, last_updated=view["last_updated"])

    def create_job(self, email, dataset_id, job_name, job_type, params):
        dataset_id = int(dataset_id)
        client = self.datastore_client
        self.__get_key_and_dataset(email, dataset_id)
        entity = datastore.Entity(client.key(JOB), exclude_from_indexes=["params"])
        import json

        entity.update(
            dict(
                dataset_id=dataset_id,
                email=email,
                name=job_name,
                type=job_type,
                status="",
                submitted=datetime.datetime.utcnow(),
                params=json.dumps(params),
            )
        )
        client.put(entity)
        job_id = entity.id
        return str(job_id)

    def get_job(self, email, job_id, return_type):
        client = self.datastore_client
        job_id = int(job_id)
        key = client.key(JOB, job_id)
        result = client.get(key)
        self.__get_key_and_dataset(email, result["dataset_id"])
        if return_type == "result":
            return result["result"]
        elif return_type == "status":
            return dict(status=result["status"])
        elif return_type == "params":
            return dict(params=result.get("params"))

    def update_job(self, email, job_id, status, result):
        if not self.capabilities()[SERVER_CAPABILITY_JOBS]:
            return
        client = self.datastore_client
        job_id = int(job_id)
        is_complete = result is not None
        key = client.key(JOB, job_id)
        entity = client.get(key)
        if not is_complete:
            self.__get_key_and_dataset(email, entity["dataset_id"])

        if is_complete:
            if os.environ.get(CIRRO_JOB_RESULTS) is not None:  # save to directory
                result = save_job_result_to_file(result, job_id)
            else:
                raise ValueError("Please specify job result location")
            entity.update(dict(result=result, status=status))
            client.put(entity)
        else:
            entity.update(dict(status=status))
            client.put(entity)

    def get_jobs(self, email, dataset_id):
        client = self.datastore_client
        results = []

        query = client.query(kind=JOB)
        dataset_id = int(dataset_id)
        self.__get_key_and_dataset(email, dataset_id, False)
        query.add_filter("dataset_id", "=", dataset_id)

        for result in query.fetch():
            results.append(
                dict(
                    id=result.id,
                    name=result["name"],
                    type=result["type"],
                    submitted=result.get("submitted"),
                    status=result.get("status"),
                    email=result.get("email"),
                )
            )
        return results

    def delete_job(self, email, job_id):
        client = self.datastore_client
        job_id = int(job_id)
        key = client.key(JOB, job_id)
        result = client.get(key)
        if result["email"] == email:
            client.delete(key)
        else:
            raise InvalidUsage("Not authorized", 403)



def mode_and_purity(x):
    value_counts = x.value_counts(sort=False)
    largest = value_counts.nlargest(1)
    purity = largest[0] / value_counts.sum()
    return largest.index[0], purity


def sparse_density(x):
    return x.sparse.density


def np_density(x):
    return (x != 0).sum() / len(x)



import os
import logging

from cirrocumulus.envir import CIRRO_LOG_LEVEL


logger = logging.getLogger("cirro")
log_level = os.environ.get(CIRRO_LOG_LEVEL, "ERROR")
logger.propagate = False
logger.setLevel(log_level)
logger.addHandler(logging.StreamHandler())  # Logs go to stderr
logger.handlers[-1].setFormatter(logging.Formatter("%(message)s"))
logger.handlers[-1].setLevel(log_level)



import sys
import math
import os.path
import argparse

import numpy as np
import anndata

import cirrocumulus.io_util
from cirrocumulus.anndata_dataset import read_adata
from cirrocumulus.util import get_fs


def concat_spatial(paths: list[str], output_path: str, ncols: int = 2):
    datasets = []
    images = []
    spot_diameters = []
    common_obsm_keys = None
    unique_dataset_names = set()
    for path in paths:
        filesystem = get_fs(path)
        adata = read_adata(path, filesystem, False)
        adata.var.index = adata.var.index.str.upper()
        dataset_name = os.path.basename(path)
        dataset_name = os.path.splitext(dataset_name)[0]
        if dataset_name.endswith("_filtered_feature_bc_matrix"):
            dataset_name = dataset_name[: -len("_filtered_feature_bc_matrix")]
        counter = 1
        while dataset_name in unique_dataset_names:
            dataset_name = dataset_name + "-" + str(counter)
            counter = counter + 1
        unique_dataset_names.add(dataset_name)
        spatial_dir = os.path.join(os.path.dirname(path), "spatial")
        if os.path.exists(spatial_dir) and cirrocumulus.io_util.add_spatial(adata, spatial_dir):
            spatial = adata.uns["images"]
            spot_diameters.append(spatial[0]["spot_diameter"])
            import imageio.v3 as iio

            images.append(iio.imread(spatial[0]["image"]))
        adata.obs["dataset"] = dataset_name
        if common_obsm_keys is None:
            common_obsm_keys = set()
            common_obsm_keys.update(adata.obsm.keys())
        else:
            common_obsm_keys.intersection(adata.obsm.keys())
        adata.obs.index = adata.obs.index + "-" + dataset_name
        datasets.append(adata)

    ncols = min(ncols, len(datasets))
    nrows = int(math.ceil(len(datasets) / ncols))
    max_width = 0
    max_height = 0
    for i in range(len(images)):
        image_shape = images[i].shape
        max_width = max(max_width, image_shape[1])
        max_height = max(max_height, image_shape[0])
    indices = []
    for i in range(nrows):
        for j in range(ncols):
            indices.append((i, j))

    do_concat_spatial = len(images) == len(datasets)
    common_obsm = dict()
    if len(common_obsm_keys) > 0:
        for key in common_obsm_keys:
            if key != "tissue_hires":
                coords_list = []
                for i in range(len(datasets)):
                    coords = datasets[i].obsm[key]
                    if coords.shape[1] in [2, 3]:
                        row_index, col_index = indices[i]
                        # 0,0 is at lower left, start at upper-left
                        row_index = nrows - row_index - 1
                        coords[:, 0] = np.interp(
                            coords[:, 0],
                            (coords[:, 0].min(), coords[:, 0].max()),
                            (col_index, col_index + 1),
                        )
                        coords[:, 1] = np.interp(
                            coords[:, 1],
                            (coords[:, 1].min(), coords[:, 1].max()),
                            (row_index, row_index + 1),
                        )
                        coords_list.append(coords)
                if len(coords_list) == len(datasets):
                    common_obsm[key] = np.concatenate(coords_list)
    if do_concat_spatial:
        combined_image = np.zeros(
            (max_height * nrows, max_width * ncols, images[0].shape[2]), dtype=images[0].dtype
        )
        combined_image[:] = images[0].max()
        adata_spatial_coords = []
        for i in range(len(datasets)):
            spatial_coords = datasets[i].obsm["tissue_hires"]  # x, y coords
            row_index, col_index = indices[i]
            if row_index > 0:
                spatial_coords[:, 1] = spatial_coords[:, 1] + max_height * row_index
            if col_index > 0:
                spatial_coords[:, 0] = spatial_coords[:, 0] + max_width * col_index
            adata_spatial_coords.append(spatial_coords)
            img = images[i]
            x_start = row_index * max_height
            y_start = col_index * max_width
            combined_image[x_start : x_start + img.shape[0], y_start : y_start + img.shape[1]] = img

    combined_adata = anndata.concat(datasets, join="outer")
    combined_adata.obs_names_make_unique()
    for key in common_obsm:
        combined_adata.obsm[key] = common_obsm[key]
    if not output_path.lower().endswith(".h5ad"):
        output_path = output_path + ".h5ad"
    output_dir = os.path.dirname(output_path)
    if output_dir == "":
        output_dir = "."
    os.makedirs(output_dir, exist_ok=True)
    if do_concat_spatial:
        combined_adata.obsm["tissue_hires"] = np.concatenate(adata_spatial_coords)

        spatial_dir = os.path.join(output_dir, "spatial")
        os.makedirs(spatial_dir, exist_ok=True)
        combined_image_path = os.path.join(spatial_dir, "tissue_hires.png")
        iio.imwrite(combined_image_path, combined_image)
        _spot_diameter = spot_diameters[0]
        spot_diameters_differ = False
        for i in range(1, len(spot_diameters)):
            if spot_diameters[i] != _spot_diameter:
                spot_diameters_differ = True
                break
        if spot_diameters_differ:  # each spot has own diameter
            _spot_diameters = []
            for i in range(len(spot_diameters)):
                per_spot_diameter = np.zeros(datasets[i].shape[0])
                per_spot_diameter[:] = spot_diameters[i]
                _spot_diameters.append(per_spot_diameter)
            spot_diameter = np.concatenate(_spot_diameters)
            assert len(spot_diameter) == combined_adata.shape[0]
        else:
            spot_diameter = spot_diameters[0]
        combined_adata.uns["images"] = dict(
            type="image",
            name="tissue_hires",
            image=combined_image_path,
            spot_diameter=spot_diameter,
        )

    combined_adata.write(output_path)


def create_parser(description=False):
    parser = argparse.ArgumentParser(
        description=(
            "Concatenate datasets in a grid layout. If all the datasets are spatial datasets, then tissue images are concatenated."
            if description
            else None
        )
    )
    parser.add_argument(
        "dataset",
        help="Paths to dataset in h5ad, 10x h5, loom, or Seurat (rds) format. For spatial dataset, a 'spatial' directory must exist in the same directory as each dataset",
        nargs="+",
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Output path to write concatenated h5ad file. A spatial directory will be created if all input datasets are spatial",
        default="data-concat.h5ad",
    )
    parser.add_argument(
        "-c",
        "--cols",
        dest="cols",
        help="Number of columns in tiled layout",
        default=2,
        type=int,
        required=False,
    )
    return parser


def main(argsv):
    parser = create_parser(True)
    args = parser.parse_args(argsv)
    concat_spatial(args.dataset, args.output, args.cols)


if __name__ == "__main__":
    main(sys.argv[1:])



import json

import h5py
import numpy as np

from cirrocumulus.abstract_backed_dataset import AbstractBackedDataset


class H5ADDataset(AbstractBackedDataset):
    def __init__(self):
        super().__init__()

    def get_suffixes(self):
        return ["h5ad"]

    def is_group(self, node):
        return isinstance(node, h5py.Group)

    def open_group(self, filesystem, path):
        return h5py.File(filesystem.open(path, "rb"), mode="r")

    def slice_dense_array(self, X, indices):
        # indexing elements must be in increasing order
        order = np.argsort(indices)
        ordered = indices[order]
        rev_order = np.argsort(order)
        value = X[:, ordered]

        return value[:, rev_order]

    def get_schema(self, filesystem, path):
        with h5py.File(filesystem.open(path, "rb"), mode="r") as f:
            return json.loads(str(f["uns"]["cirro-schema"][...].astype(str)))



import numpy as np
import pandas as pd
import anndata
from pandas import CategoricalDtype


DATA_TYPE_MODULE = "module"
DATA_TYPE_UNS_KEY = "data_type"
ADATA_MODULE_UNS_KEY = "anndata_module"
ADATA_LAYERS_UNS_KEY = "anndata_layers"


def get_base(adata):
    base = None
    if "log1p" in adata.uns and adata.uns["log1p"]["base"] is not None:
        base = adata.uns["log1p"][base]
    return base


def adata_to_df(adata):
    df = pd.DataFrame(adata.X, index=adata.obs.index, columns=adata.var.index)
    for key in adata.layers.keys():
        df2 = pd.DataFrame(
            adata.layers[key],
            index=adata.obs.index.astype(str) + "-{}".format(key),
            columns=adata.var.index,
        )
        df = pd.concat((df, df2), axis=0)

    df = df.T.join(adata.var)
    df.index.name = "id"
    return df.reset_index()


def get_scanpy_marker_keys(dataset):
    marker_keys = []
    try:
        dataset.uns  # dataset can be AnnData or zarr group
    except AttributeError:
        return marker_keys
    from collections.abc import Mapping

    for key in dataset.uns.keys():
        rank_genes_groups = dataset.uns[key]
        if (
            isinstance(rank_genes_groups, Mapping)
            and "names" in rank_genes_groups
            and (
                "pvals" in rank_genes_groups
                or "pvals_adj" in rank_genes_groups
                or "scores" in rank_genes_groups
            )
            and len(rank_genes_groups["names"][0]) > 0
            and not isinstance(rank_genes_groups["names"][0][0], bytes)
        ):
            marker_keys.append(key)
    return marker_keys


def get_pegasus_marker_keys(dataset):
    marker_keys = []
    try:
        dataset.varm  # dataset can be AnnData or zarr group
    except AttributeError:
        return marker_keys
    for key in dataset.varm.keys():
        d = dataset.varm[key]
        if isinstance(d, np.recarray):
            try:
                "".join(d.dtype.names).index("log2Mean")
                marker_keys.append(key)
            except ValueError:
                pass
    return marker_keys


def obs_stats(adata, columns):
    df = adata.obs[columns]
    # variables on columns, stats on rows, transpose so that stats are on columns
    return df.agg(["min", "max", "sum", "mean"]).T


def X_stats(adata):
    X = adata.X
    return pd.DataFrame(
        data={
            "min": X.min(axis=0).toarray().flatten(),
            "max": X.max(axis=0).toarray().flatten(),
            "sum": X.sum(axis=0).flatten(),
            "numExpressed": X.getnnz(axis=0),
            "mean": X.mean(axis=0),
        },
        index=adata.var.index,
    )


def dataset_schema(dataset, n_features=10):
    """Gets dataset schema.

    Returns
        schema dict. Example:
        {"version":"1.0.0",
        "categoryOrder":{
            "louvain":["0","1","2","3","4","5","6","7"],
            "leiden":["0","1","2","3","4","5","6","7"]},
        "var":["TNFRSF4","CPSF3L","ATAD3C"],
        "obs":["percent_mito","n_counts"],
        "obsCat":["louvain","leiden"],
        "shape":[2638,1838],
        "embeddings":[{"name":"X_pca","dimensions":3},{"name":"X_pca","dimensions":2},{"name":"X_umap","dimensions":2}]
    }
    """
    obs_cat = []
    obs = []
    marker_results = []
    de_results_format = "records"

    prior_marker_results = dataset.uns.get("markers", [])
    if isinstance(prior_marker_results, str):
        import json

        prior_marker_results = json.loads(prior_marker_results)
    marker_results += prior_marker_results

    de_results = []  # array of dicts containing params logfoldchanges, pvals_adj, scores, names
    category_to_order = {}
    embeddings = []
    field_to_value_to_color = dict()  # field -> value -> color

    for scanpy_marker_key in get_scanpy_marker_keys(dataset):
        rank_genes_groups = dataset.uns[scanpy_marker_key]
        has_fc = "logfoldchanges" in rank_genes_groups
        min_fold_change = 1
        params = rank_genes_groups["params"]
        if not isinstance(params, dict):
            from anndata._io.zarr import read_attribute

            params = {k: read_attribute(params[k]) for k in params.keys()}

        # pts and pts_rest in later scanpy versions
        rank_genes_groups_keys = list(rank_genes_groups.keys())
        for k in ["params", "names"]:
            if k in rank_genes_groups_keys:
                rank_genes_groups_keys.remove(k)
        if "pvals" in rank_genes_groups_keys and "pvals_adj" in rank_genes_groups_keys:
            rank_genes_groups_keys.remove("pvals")
        category = "{} ({})".format(params["groupby"], scanpy_marker_key)
        de_result_name = category
        de_result_df = None
        group_names = rank_genes_groups["names"].dtype.names
        de_result = dict(
            id="cirro-{}".format(scanpy_marker_key),
            type="de",
            readonly=True,
            groups=group_names,
            fields=rank_genes_groups_keys,
            name=de_result_name,
        )
        for group_name in group_names:
            group_df = pd.DataFrame(index=rank_genes_groups["names"][group_name][...])
            group_df = group_df[group_df.index != "nan"]
            for rank_genes_groups_key in rank_genes_groups_keys:
                values = rank_genes_groups[rank_genes_groups_key][group_name][...]
                column_name = "{}:{}".format(group_name, rank_genes_groups_key)
                group_df[column_name] = values

            if de_result_df is None:
                de_result_df = group_df
            else:
                de_result_df = de_result_df.join(group_df, how="outer")
            if n_features > 0:
                markers_df = group_df
                if has_fc:
                    markers_df = group_df[
                        group_df["{}:logfoldchanges".format(group_name)] > min_fold_change
                    ]
                if len(markers_df) > 0:
                    marker_results.append(
                        dict(
                            category=de_result_name,
                            name=str(group_name),
                            features=markers_df.index[:n_features],
                        )
                    )

        if de_results_format == "records":
            de_result_data = de_result_df.reset_index().to_dict(orient="records")
        else:
            de_result_data = dict(index=de_result_df.index[...])
            for c in de_result_df:
                de_result_data[c] = de_result_df[c]

        de_result["params"] = params
        de_result["data"] = de_result_data
        de_results.append(de_result)
    for pg_marker_key in get_pegasus_marker_keys(dataset):
        de_res = dataset.varm[pg_marker_key]
        key_name = pg_marker_key
        if pg_marker_key.startswith("rank_genes_"):
            key_name = pg_marker_key[len("rank_genes_") :]
        names = de_res.dtype.names
        field_names = set()  # e.g. 1:auroc
        group_names = set()
        for name in names:
            index = name.rindex(":")
            field_name = name[index + 1 :]
            group_name = name[:index]
            field_names.add(field_name)
            group_names.add(group_name)
        group_names = list(group_names)
        field_names = list(field_names)

        de_result_df = pd.DataFrame(data=de_res, index=dataset.var.index)
        de_result_df.index.name = "index"
        if de_results_format == "records":
            de_result_data = de_result_df.reset_index().to_dict(orient="records")
        else:
            de_result_data = dict(index=de_result_df.index)
            for c in de_res:
                de_result_data[c] = de_result_df[c]
        de_result = dict(
            id="cirro-{}".format(pg_marker_key),
            type="de",
            name=key_name,
            color="log2FC" if "log2FC" in field_names else field_names[0],
            size="mwu_qval" if "mwu_qval" in field_names else field_names[0],
            groups=group_names,
            fields=field_names,
        )
        de_result["data"] = de_result_data
        de_results.append(de_result)

        if n_features > 0:
            field_use = None
            for field in ["mwu_qval", "auroc", "t_qval"]:
                if field in field_names:
                    field_use = field
                    break
            if field_use is not None:
                field_ascending = field_use != "auroc"
                for group_name in group_names:
                    fc_column = "{}:log2FC".format(group_name)
                    name = "{}:{}".format(group_name, field_name)

                    idx_up = de_result_df[fc_column] > 0
                    df_up = de_result_df.loc[idx_up].sort_values(
                        by=[name, fc_column], ascending=[field_ascending, False]
                    )
                    features = df_up[:n_features].index.values

                    marker_results.append(
                        dict(category=key_name, name=str(group_name), features=features)
                    )

    categories_node = dataset.obs["__categories"] if "__categories" in dataset.obs else None
    for key in dataset.obs.keys():
        if categories_node is not None and (key == "__categories" or key == "index"):
            continue
        val = dataset.obs[key]
        if categories_node is not None and key in categories_node:
            categories = categories_node[key][...]
            ordered = categories_node[key].attrs.get("ordered", False)
            val = pd.Categorical.from_codes(val[...], categories, ordered=ordered)

        if (
            isinstance(val.dtype, CategoricalDtype)
            or pd.api.types.is_bool_dtype(val)
            or pd.api.types.is_object_dtype(val)
        ):
            obs_cat.append(key)
        else:
            obs.append(key)
        if isinstance(val.dtype, CategoricalDtype):
            categories = val.cat.categories
            if len(categories) < 100:  # preserve order
                category_to_order[key] = dataset.obs[key].cat.categories

            color_field = key + "_colors"
            if color_field in dataset.uns:
                colors = dataset.uns[color_field][...].flatten()

                if len(categories) == len(colors):
                    color_map = dict()
                    for j in range(len(categories)):
                        color_map[str(categories[j])] = colors[j]
                    field_to_value_to_color[key] = color_map
                else:
                    print("Invalid number of categories for {}".format(color_field))

    # spatial_node = adata.uns['spatial'] if 'spatial' in adata.uns else None
    #
    # if spatial_node is not None:
    #     spatial_node_keys = list(spatial_node.keys())  # list of datasets
    #     if len(spatial_node_keys) == 1:
    #         spatial_node = spatial_node[spatial_node_keys[0]]

    images_node = dataset.uns.get(
        "images", []
    )  # list of {type:image or meta_image, name:image name, image:path to image, spot_diameter:Number}
    image_names = list(map(lambda x: x["name"], images_node))
    layers = []
    try:
        dataset.layers  # dataset can be AnnData or zarr group
        layers = list(dataset.layers.keys())
    except AttributeError:
        pass

    for key in dataset.obsm.keys():
        dim = dataset.obsm[key].shape[1]
        if 1 < dim <= 3:
            embedding = dict(name=key, dimensions=dim)
            if dim == 2:
                try:
                    image_index = image_names.index(key)
                    embedding["spatial"] = images_node[image_index]
                except ValueError:
                    pass

            embeddings.append(embedding)
    meta_images = dataset.uns.get("meta_images", [])
    for meta_image in meta_images:
        embeddings.append(meta_image)
    schema_dict = {"version": "1.0.0"}
    schema_dict["results"] = de_results
    schema_dict["colors"] = field_to_value_to_color
    schema_dict["markers"] = marker_results
    schema_dict["embeddings"] = embeddings
    schema_dict["categoryOrder"] = category_to_order
    schema_dict["layers"] = layers
    schema_dict["timepoint_field"] = dataset.uns.get("timepoint_field", "day")
    var_df = dataset.var
    if not isinstance(var_df, pd.DataFrame):
        from anndata._io.zarr import read_attribute

        var_df = read_attribute(dataset.var)
    var_df.index.name = "id"
    schema_dict["var"] = var_df.reset_index().to_dict(orient="records")
    modules_df = None
    if ADATA_MODULE_UNS_KEY in dataset.uns and isinstance(
        dataset.uns[ADATA_MODULE_UNS_KEY], anndata.AnnData
    ):
        modules_df = dataset.uns[ADATA_MODULE_UNS_KEY].var
        #  if not isinstance(module_var, pd.DataFrame):
        #             from anndata._io.zarr import read_attribute
        #             module_var = read_attribute(module_var)
    if modules_df is not None:
        modules_df.index.name = "id"
        schema_dict["modules"] = modules_df.reset_index().to_dict(orient="records")

    schema_dict["obs"] = obs
    schema_dict["obsCat"] = obs_cat
    shape = dataset.shape if isinstance(dataset, anndata.AnnData) else dataset.X.attrs.shape
    schema_dict["shape"] = shape
    return schema_dict



import re

import h5py
import numpy as np
import fsspec
import pandas as pd
import anndata

from cirrocumulus.ot.population import Population
from cirrocumulus.ot.util import chain_transport_maps, find_path, unique_timepoint
from cirrocumulus.util import get_scheme


class TransportMapModel:
    def __init__(self, tmaps, meta, timepoints=None, day_pairs=None, cache=False):
        self.tmaps = tmaps
        self.meta = meta
        self.cache = cache
        if timepoints is None:
            timepoints = sorted(meta["day"].unique())
        self.timepoints = timepoints

        if day_pairs is None:
            day_pairs = [(timepoints[i], timepoints[i + 1]) for i in range(len(timepoints) - 1)]
        self.day_pairs = day_pairs

    def trajectories(self, populations):
        """Computes a trajectory for each population.

        Parameters
        ----------
        self : wot.TransportMapModel
            The TransportMapModel used to find ancestors and descendants of the population
        populations : list of wot.Population
            The target populations such as ones from self.population_from_cell_sets. THe populations must be from the same time.

        Returns
        -------
        trajectories : anndata.AnnData
            Rows : all cells, Columns : populations index. At point (i, j) : the probability that cell i is an
            ancestor/descendant of population j
        """
        unique_timepoint(*populations)  # check for unique timepoint
        trajectories = []

        populations = Population.copy(*populations, normalize=True, add_missing=False)
        population_names = [p.name for p in populations]
        initial_populations = populations

        def update(head, populations_to_update):
            idx = 0 if head else len(trajectories)
            trajectories.insert(idx, np.array([pop.p for pop in populations_to_update]).T)

        update(True, populations)
        while self.can_pull_back(*populations):
            populations = self.pull_back(*populations, as_list=True)
            update(True, populations)
        populations = initial_populations
        while self.can_push_forward(*populations):
            populations = self.push_forward(*populations, as_list=True)
            update(False, populations)

        X = np.concatenate(trajectories)
        return anndata.AnnData(
            X=X,
            obs=self.meta.copy(),
            var=pd.DataFrame(index=population_names),
        )

    def get_coupling(self, t0, t1, covariate=None):
        """Loads a coupling for a given pair of timepoints.

        Parameters
        ----------
        t0 : int or float
            Source timepoint of the transport map.
        t1 : int of float
            Destination timepoint of the transport map.
        covariate : None or (int, int), optional
            Restrict to certain covariate values. Do not restrict if None

        Returns
        -------
        tmap : anndata.AnnData
            The transport map from t0 to t1
        """
        if t0 not in self.timepoints or t1 not in self.timepoints:
            raise ValueError("Day pair {}, {} not found".format(t0, t1))

        atomic = (
            self.day_pairs is not None and (t0, t1) in self.day_pairs
        ) or self.timepoints.index(t1) == self.timepoints.index(t0) + 1

        if not atomic and covariate is not None:
            raise ValueError("Covariate-restricted transport maps can only be atomic")

        if atomic:
            if covariate is None:
                key = (t0, t1)
            else:
                cv0, cv1 = covariate
                key = (t0, t1, str(cv0), str(cv1))
            ds_or_path = self.tmaps.get(key)
            if ds_or_path is None:
                raise ValueError("No transport map found for {}", key)
            if type(ds_or_path) is anndata.AnnData:
                return ds_or_path
            with fsspec.open(ds_or_path) as f:
                ds = anndata.read(f)
            if self.cache:
                self.tmaps[key] = ds
            return ds

        else:
            path = find_path(t0, t1, self.day_pairs, self.timepoints)
            return chain_transport_maps(self, path)

    def can_push_forward(self, *populations):
        """Checks if the populations can be pushed forward.

        Parameters
        ----------
        *populations : wot.Population
            Measure over the cells at a given timepoint to be pushed forward.

        Returns
        -------
        result : bool
            True if the populations can be pushed forward

        Raises
        ------
        ValueError
            If all populations are not in the same timepoint
        """
        return self.timepoints.index(unique_timepoint(*populations)) < len(self.timepoints) - 1

    def can_pull_back(self, *populations):
        """Checks if the populations can be pulled back.

        Parameters
        ----------
        *populations : wot.Population
            Measure over the cells at a given timepoint to be pulled back.

        Returns
        -------
        result : bool
            True if the populations can be pulled back.

        Raises
        ------
        ValueError
            If all populations are not in the same timepoint
        """
        return self.timepoints.index(unique_timepoint(*populations)) > 0

    def push_forward(self, *populations, to_time=None, normalize=True, as_list=False):
        """Pushes the population forward through the computed transport maps.

        Parameters
        ----------
        *populations : wot.Population
            Measure over the cells at a given timepoint to be pushed forward.
        to_time : int or float, optional
            Destination timepoint to push forward to.
        normalize : bool, optional, default: True
            Wether to normalize to a probability distribution or keep growth.
        as_list : bool, optional, default: False
            Wether to return a list of length 1 when a single element is passed, or a Population

        Returns
        -------
        result : wot.Population
            The push forward of the input population through the proper transport map.
            Array of populations if several populations were given as input.

        Raises
        ------
        ValueError
            If there is no further timepoint to push the population forward.
        ValueError
            If several populations are given as input but dot live in the same timepoint.

        Examples
        --------
        >>> self.push_forward(pop, to_time = 2) # -> wot.Population
        Pushing several populations at once
        >>> self.push_forward(pop1, pop2, pop3) # -> list of wot.Population
        Pulling back after pushing forward
        >>> self.pull_back(self.push_forward(pop))
        Same, but several populations at once
        >>> self.pull_back(* self.push_forward(pop1, pop2, pop3))
        """
        i = self.timepoints.index(unique_timepoint(*populations))
        j = i + 1 if to_time is None else self.timepoints.index(to_time)

        if i == -1:
            raise ValueError("Timepoint not found")
        if j == -1:
            raise ValueError("Destination timepoint not found")
        if j >= len(self.timepoints):
            raise ValueError("No further timepoints. Unable to push forward")
        if i > j:
            raise ValueError("Destination timepoint is before source. Unable to push forward")

        p = np.vstack([pop.p for pop in populations])
        while i < j:
            t0 = self.timepoints[i]
            t1 = self.timepoints[i + 1]
            tmap = self.get_coupling(t0, t1)
            p = p @ tmap.X
            if normalize:
                p = (p.T / np.sum(p, axis=1)).T
            i += 1

        result = [Population(self.timepoints[i], p[k, :]) for k in range(p.shape[0])]
        if len(result) == 1 and not as_list:
            return result[0]
        else:
            return result

    def pull_back(self, *populations, to_time=None, normalize=True, as_list=False):
        """Pulls the population back through the computed transport maps.

        Parameters
        ----------
        *populations : wot.Population
            Measure over the cells at a given timepoint to be pushed forward.
        to_time : int or float, optional
            Destination timepoint to pull back to.
        normalize : bool, optional, default: True
            Wether to normalize to a probability distribution or keep growth.
        as_list : bool, optional, default: False
            Wether to return a listof length 1 when a single element is passed, or a Population

        Returns
        -------
        result : wot.Population
            The pull back of the input population through the proper transport map.
            Array of populations if several populations were given as input.

        Raises
        ------
        ValueError
            If there is no previous timepoint to pull the population back.
        ValueError
            If several populations are given as input but dot live in the same timepoint.

        Examples
        --------
        >>> self.pull_back(pop, to_time = 0) # -> wot.Population
        Pushing several populations at once
        >>> self.pull_back(pop1, pop2, pop3) # -> list of wot.Population
        Pulling back after pushing forward
        >>> self.pull_back(self.push_forward(pop))
        Same, but several populations at once
        >>> self.pull_back(* self.push_forward(pop1, pop2, pop3))
        """
        i = self.timepoints.index(unique_timepoint(*populations))
        j = i - 1 if to_time is None else self.timepoints.index(to_time)

        if i == -1:
            raise ValueError("Timepoint not found")
        if i == 0:
            raise ValueError("No previous timepoints. Unable to pull back")
        if j == -1:
            raise ValueError("Destination timepoint not found")
        if i < j:
            raise ValueError("Destination timepoint is after source. Unable to pull back")

        p = np.vstack([pop.p for pop in populations])
        while i > j:
            t1 = self.timepoints[i]
            t0 = self.timepoints[i - 1]
            tmap = self.get_coupling(t0, t1)
            p = (tmap.X @ p.T).T
            if normalize:
                p = (p.T / np.sum(p, axis=1)).T
            i -= 1

        result = [Population(self.timepoints[i], p[k, :]) for k in range(p.shape[0])]
        if len(result) == 1 and not as_list:
            return result[0]
        else:
            return result

    def population_from_ids(self, *ids, at_time):
        """Constructs a population uniformly distributed among the ids given as input.

        Parameters
        ----------
        *ids : list of str
            The list of cell ids that belong to that population.
        at_time : int or float
            The time at which to construct the population.
            Cells that come from a different time point will be ignored.

        Returns
        -------
        *populations : list of wot.Population
            A population, uniformly distributed over the cells given as input.
            Returns None if the generated population would be empty



        Examples
        --------
        >>> cell_set = [ 'cell_1', 'cell_2', 'cell_3' ]
        >>> self.population_from_ids(cell_set) # -> wot.Population
        Multiple populations at once
        >>> multi_cell_sets = {
        >>>   'set_a': [ 'cell_a1', 'cell_a2'],
        >>>   'set_b': [ 'cell_b1', 'cell_b2'],
        >>> }
        >>> self.population_from_ids(* multi_cell_sets.values()) # -> list of wot.Population

        Notes
        -----
        The Population class is a measure over the cells at a given timepoint.
        It does not necessarily sum to 1. However, this method always returns a probability distribution over the cells of that time point.
        """
        day = float(at_time)
        df = self.meta[self.meta["day"] == day]

        def get_population(ids_el):
            cell_indices = df.index.get_indexer_for(ids_el)
            cell_indices = cell_indices[cell_indices > -1]

            if len(cell_indices) == 0:
                return None
            p = np.zeros(len(df), dtype=np.float64)
            p[cell_indices] = 1.0

            return Population(day, p)

        result = [get_population(ids_el) for ids_el in ids]
        return result


def read_transport_map_dir(transport_map_url, with_covariates=False, cache=False):
    tmaps = {}  # maps day pair to transport map url

    day_regex = "([0-9]*\\.?[0-9]+)"
    tmap_prefix = ".*"
    if with_covariates:
        pattern = re.compile(
            tmap_prefix
            + "_{}_{}_cv([a-zA-Z0-9]+)_cv([a-zA-Z0-9]+)[\\.h5ad|\\.loom|\\.zarr]".format(
                day_regex, day_regex
            )
        )
    else:
        pattern = re.compile(
            tmap_prefix + "_{}_{}[\\.h5ad|\\.loom|\\.zarr]".format(day_regex, day_regex)
        )
    scheme = get_scheme(transport_map_url)
    fs = fsspec.filesystem(scheme)

    for path in fs.glob(transport_map_url + "/*[.zarr,.h5ad]"):
        m = pattern.match(path)
        if scheme != "file":
            path = scheme + "://" + path
        if m is not None:
            try:
                t1 = float(m.group(1))
                t2 = float(m.group(2))
                if with_covariates:
                    cv1 = m.group(3)
                    cv2 = m.group(4)
                    tmaps[(t1, t2, cv1, cv2)] = path
                else:
                    tmaps[(t1, t2)] = path
            except ValueError:
                print("Unable to find day pair for " + path)
                pass

    if len(tmaps) == 0:
        raise ValueError("No transport maps found")
    day_pairs = set()
    timepoints = set()
    tmap_keys = list(tmaps.keys())
    tmap_keys.sort(key=lambda x: x[0])
    meta = None
    for i in range(len(tmap_keys)):
        key = tmap_keys[i]
        t0 = key[0]
        t1 = key[1]
        day_pairs.add((t0, t1))
        timepoints.add(t0)
        timepoints.add(t1)
        if not with_covariates:
            path = tmaps[key]
            if path.endswith(".zarr"):
                raise ValueError("Not yet implemented")
            else:
                with fs.open(path) as stream:
                    with h5py.File(stream, "r") as f:
                        # only read column ids if the last timepoint
                        if path.endswith(".loom"):
                            rids = f["/row_attrs/id"][()].astype(str)
                            cids = (
                                f["/col_attrs/id"][()].astype(str)
                                if i == len(tmap_keys) - 1
                                else None
                            )
                        else:
                            obs = f["/obs"]
                            var = f["/var"]
                            obs_key = obs.attrs.get("_index", "index")
                            var_key = var.attrs.get("_index", "index")
                            rids = obs[obs_key][()].astype(str)
                            cids = var[var_key][()].astype(str) if i == len(tmap_keys) - 1 else None
                        rdf = pd.DataFrame(index=rids, data={"day": t0})
                        cdf = (
                            pd.DataFrame(index=cids, data={"day": t1}) if cids is not None else None
                        )
                        if meta is None:
                            meta = rdf if cdf is None else pd.concat((rdf, cdf), copy=False)
                        else:
                            meta = (
                                pd.concat((meta, rdf), copy=False)
                                if cdf is None
                                else pd.concat((meta, rdf, cdf), copy=False)
                            )

    timepoints = sorted(timepoints)
    return TransportMapModel(
        tmaps=tmaps, meta=meta, timepoints=timepoints, day_pairs=day_pairs, cache=cache
    )



import pandas as pd
import scipy.sparse
from pandas import CategoricalDtype

from cirrocumulus.anndata_util import ADATA_LAYERS_UNS_KEY, ADATA_MODULE_UNS_KEY
from cirrocumulus.dotplot_aggregator import DotPlotAggregator
from cirrocumulus.feature_aggregator import FeatureAggregator
from cirrocumulus.ids_aggregator import IdsAggregator
from cirrocumulus.unique_aggregator import UniqueAggregator


def get_mask(dataset_api, dataset, dataset_info, data_filters):
    measures = set()
    dimensions = set()
    basis = set()
    for data_filter in data_filters:
        _measures, _dimensions, _basis = data_filter_keys(data_filter, dataset_info)
        measures.update(_measures)
        dimensions.update(_dimensions)
        basis.update(_basis)
    keys = get_type_to_measures(measures)
    keys["obs"] = list(dimensions)
    keys["basis"] = list(basis)
    adata = dataset_api.read_dataset(keys=keys, dataset=dataset)
    result = []
    for data_filter in data_filters:
        result.append(get_filter_expr(adata, data_filter))
    return result, adata


def apply_filter(adata, data_filter):
    keep_expr = get_filter_expr(adata, data_filter)
    return adata[keep_expr] if keep_expr is not None else adata


def get_filter_str(data_filter):
    user_filters = data_filter.get("filters", [])
    # example
    # {
    #   "filters": [
    #     {
    #       "field": "leiden",
    #       "operation": "in",
    #       "value": [
    #         "0"
    #       ]
    #     }
    #   ],
    #   "combine": "and"
    # }
    #    combine_filters = data_filter.get("combine", "and")
    s = []
    for filter_obj in user_filters:
        field = filter_obj["field"]
        op = filter_obj["operation"]
        value = filter_obj["value"]
        if op == "in":
            op = ""
        if not isinstance(field, dict) and not field == "__index":
            if isinstance(value, list):
                tmp = []
                for v in value:
                    tmp.append(str(v))
                value = "_".join(tmp)
                s.append(field + "-" + value)
            else:
                return None
        else:
            return None
    return "".join(s).replace("/", "-").replace(" ", "-").replace("\\", "-")


def _flatten_array(val):
    if scipy.sparse.issparse(val):
        val = val.toarray().flatten()
    if hasattr(val, "sparse"):
        val = val.sparse.to_dense()
    if isinstance(val, pd.Series):
        val = val.values
    return val


def get_filter_expr(adata, data_filter):
    keep_expr = None
    if data_filter is not None:
        user_filters = data_filter.get("filters", [])
        combine_filters = data_filter.get("combine", "and")
        for filter_obj in user_filters:
            field = filter_obj["field"]
            op = filter_obj["operation"]
            value = filter_obj["value"]
            invert = filter_obj.get("invert", False)
            if isinstance(field, dict):  # selection box
                # selected_points_basis = get_basis(field['basis'], field.get('nbins'),
                #                                   field.get('agg'), field.get('ndim', '2'),
                #                                   field.get('precomputed', False))

                if "points" in value:
                    p = value["points"]
                    # list of indices
                    keep = pd.RangeIndex(adata.shape[0]).isin(p)

                # else:
                #     keep = None
                #     for p in value['path']:
                #         if 'z' in p:  # 3d
                #             selection_keep = \
                #                 (df[selected_points_basis['coordinate_columns'][0]] >= p['x']) & \
                #                 (df[selected_points_basis['coordinate_columns'][0]] <= p['x'] + p['width']) & \
                #                 (df[selected_points_basis['coordinate_columns'][1]] >= p['y']) & \
                #                 (df[selected_points_basis['coordinate_columns'][1]] <= p['y'] + p[
                #                     'height']) & \
                #                 (df[selected_points_basis['coordinate_columns'][2]] >= p['z']) & \
                #                 (df[selected_points_basis['coordinate_columns'][2]] <= p['z'] + p['depth'])
                #         else:
                #             selection_keep = \
                #                 (df[selected_points_basis['coordinate_columns'][0]] >= p['x']) & \
                #                 (df[selected_points_basis['coordinate_columns'][0]] <= p['x'] + p['width']) & \
                #                 (df[selected_points_basis['coordinate_columns'][1]] >= p['y']) & \
                #                 (df[selected_points_basis['coordinate_columns'][1]] <= p['y'] + p['height'])

                # keep = selection_keep | keep if keep is not None else selection_keep

            elif field == "__index":
                import numpy as np

                keep = np.zeros(adata.shape[0], dtype=bool)
                keep[value] = True
            else:
                series = (
                    adata.obs[field]
                    if field in adata.obs
                    else adata[:, adata.var.index == field].X[:, 0]
                )
                if op == "in":
                    keep = (series.isin(value)).values
                else:  # array of operations e.g. >, <
                    keep = None
                    for i in range(len(op)):
                        op_i = op[i]
                        value_i = value[i]
                        if op_i == ">":
                            keep_i = series > value_i
                        elif op_i == "=":
                            keep_i = series == value_i
                        elif op_i == "<":
                            keep_i = series < value_i
                        elif op_i == "!=":
                            keep_i = series != value_i
                        elif op_i == ">=":
                            keep_i = series >= value_i
                        elif op_i == "<=":
                            keep_i = series <= value_i
                        else:
                            raise ValueError(
                                "Unknown filter, field: {}, operation: {}, value: {}".format(
                                    field, op, value
                                )
                            )
                        keep_i = _flatten_array(keep_i)
                        keep = keep_i & keep if keep is not None else keep_i
            keep = _flatten_array(keep)
            if invert:
                keep = ~keep
            if keep_expr is not None:
                if combine_filters == "and":
                    keep_expr = keep_expr & keep
                else:
                    keep_expr = keep_expr | keep
            else:
                keep_expr = keep

    return keep_expr


def precomputed_summary(dataset_api, dataset, obs_measures, var_measures, dimensions):
    if "__count" in var_measures:
        var_measures.remove("__count")
    return dataset_api.read_precomputed_stats(
        dataset, obs_keys=dimensions + obs_measures, var_keys=var_measures
    )


def precomputed_grouped_stats(dataset_api, dataset, var_measures, dimensions):
    if (len(var_measures)) > 0 and len(dimensions) > 0:
        return dataset_api.read_precomputed_grouped_stats(
            dataset, var_keys=var_measures, obs_keys=dimensions
        )
    return []


def precomputed_embedding(dataset_api, dataset, basis, obs_measures, var_measures, dimensions):
    if (len(obs_measures) + len(var_measures) + len(dimensions)) == 0:
        obs_measures = ["__count"]
    return dataset_api.read_precomputed_basis(
        dataset, obs_keys=obs_measures + dimensions, var_keys=var_measures, basis=basis
    )


def get_var_name_type(key, default_type="X", dataset_info=None):
    index = key.find("/")
    if index == -1:
        if dataset_info is not None:
            key_type = "X" if key in dataset_info["var"] else "obs"
            return key, key_type
        return key, default_type
    else:
        key_type = key[0:index]
        name = key[index + 1 :]
        return name, key_type


def get_type_to_measures(measures):
    type2measures = dict(X=[], obs=[])
    for measure in measures:
        name, key_type = get_var_name_type(measure)
        type_measures = type2measures.get(key_type)
        if type_measures is None:
            type_measures = []
            type2measures[key_type] = type_measures
        type_measures.append(name)
    return type2measures


def check_bin_input(nbins):
    if nbins is not None:
        nbins = int(nbins)
        nbins = min(1000, nbins)
        if nbins <= 0:
            nbins = None
    return nbins


def handle_export_dataset_filters(dataset_api, dataset, data_filters):
    import json

    reformatted_filters = []
    filter_names = []
    for data_filter_obj in data_filters:
        filter_value = json.loads(data_filter_obj["value"])
        filter_names.append(data_filter_obj["name"])
        reformatted_filters.append(filter_value)

    df = get_adata(dataset_api, dataset, measures=["obs/index"], data_filters=reformatted_filters)
    result_df = pd.DataFrame(index=df["index"])
    for i in range(len(reformatted_filters)):
        data_filter = reformatted_filters[i]
        filter_name = filter_names[i]
        df_filtered = apply_filter(df, data_filter)
        df = pd.DataFrame(index=df_filtered["index"])
        df[filter_name] = True
        result_df = result_df.join(df, rsuffix="r")
    result_df.fillna(False, inplace=True)
    return result_df.to_csv()


# embedding - list of basis and coords (whether to return coordinates). For binned embeddings, also measures and dimensions.
def handle_data(
    dataset_api,
    dataset,
    embedding_list=None,
    values=None,
    grouped_stats=None,
    stats=None,
    selection=None,
):
    dimensions = set()
    measures = set()
    basis_keys = set()
    if embedding_list is not None:
        for embedding in embedding_list:
            basis_keys.add(embedding["name"])

    if values is not None:
        dimensions.update(values.get("dimensions", []))
        measures.update(values.get("measures", []))

    if selection is not None:
        data_filter = selection.get("filter")
        var_keys_filter, obs_keys_filter, selected_points_filter_basis_list = data_filter_keys(
            data_filter
        )
        selection["basis"] = selected_points_filter_basis_list
        measures.update(var_keys_filter)
        dimensions.update(obs_keys_filter)
        dimensions.update(selection.get("dimensions", []))
        measures.update(selection.get("measures", []))
        selection_embeddings = selection.get("embeddings", [])

        for embedding in selected_points_filter_basis_list + selection_embeddings:
            basis_keys.add(embedding["name"])

    if grouped_stats is not None:
        grouped_stats_dimensions = grouped_stats.get("dimensions", [])
        for d in grouped_stats_dimensions:
            if isinstance(d, list):
                dimensions.update(d)
            else:
                dimensions.add(d)
        measures.update(grouped_stats.get("measures", []))
    if stats is not None:
        dimensions.update(stats.get("dimensions", []))
        measures.update(stats.get("measures", []))

    keys = get_type_to_measures(measures)
    keys["obs"] += list(dimensions)
    keys["basis"] = list(basis_keys)
    adata = dataset_api.read_dataset(dataset=dataset, keys=keys)
    results = {}
    if values is not None:
        dimensions = values.get("dimensions", [])
        measures = values.get("measures", [])
        # measure can be X or layers
        type2measures = get_type_to_measures(measures)
        results["values"] = {}
        for key in type2measures["obs"] + dimensions:
            series = adata.obs[key]
            results["values"][key] = series
            if isinstance(series.dtype, CategoricalDtype):
                results["values"][key] = dict(
                    values=series.values.codes, categories=series.cat.categories.values
                )
            else:
                results["values"][key] = series

        def array_to_json(d, var_index, result):
            is_sparse = scipy.sparse.issparse(d)
            for i in range(len(var_index)):
                x = d[:, i]
                if is_sparse:
                    indices = x.indices
                    data = x.data
                    result[var_index[i]] = dict(indices=indices, values=data)
                else:
                    result[var_index[i]] = x

        if adata.uns.get(ADATA_MODULE_UNS_KEY) is not None:
            adata_modules = adata.uns[ADATA_MODULE_UNS_KEY]
            array_to_json(adata_modules.X, adata_modules.var.index, results["values"])

        if adata.X is not None:
            array_to_json(adata.X, adata.var.index, results["values"])
        if ADATA_LAYERS_UNS_KEY in adata.uns:
            for layer_name in adata.uns[ADATA_LAYERS_UNS_KEY].keys():
                if "layers" not in results:
                    results["layers"] = {}
                results["layers"][layer_name] = {}
                adata_layer = adata.uns[ADATA_LAYERS_UNS_KEY][layer_name]
                array_to_json(adata_layer.X, adata_layer.var.index, results["layers"][layer_name])

    if embedding_list is not None and len(embedding_list) > 0:
        results["embeddings"] = []
        for key in adata.obsm.keys():
            m = adata.obsm[key]
            ndim = m.shape[1]
            coordinates = dict()
            embedding = dict(name=key, dimensions=ndim, coordinates=coordinates)
            results["embeddings"].append(embedding)
            for i in range(ndim):
                coordinates["{}_{}".format(key, i + 1)] = m[:, i]

    if grouped_stats is not None:
        results["distribution"] = DotPlotAggregator(
            var_measures=grouped_stats.get("measures", []),
            dimensions=grouped_stats.get("dimensions", []),
        ).execute(adata)
    if stats is not None:
        dimensions = stats.get("dimensions", [])
        measures = stats.get("measures", [])
        type2measures = get_type_to_measures(measures)
        results["summary"] = FeatureAggregator(
            type2measures["obs"], type2measures["X"], dimensions
        ).execute(adata)
    if selection is not None:
        results["selection"] = {}
        dimensions = selection.get("dimensions", [])
        measures = selection.get("measures", [])
        type2measures = get_type_to_measures(measures)
        # basis_list = selection.get('basis', [])
        selection_embeddings = selection.get("embeddings", [])
        df = apply_filter(adata, data_filter)
        if len(selection_embeddings) > 0:
            results["selection"]["coordinates"] = {}
            for embedding in selection_embeddings:
                results["selection"]["coordinates"][embedding["name"]] = UniqueAggregator(
                    "index"
                ).execute(df)
        var_measures = type2measures["X"]
        results["selection"]["summary"] = FeatureAggregator(
            type2measures["obs"], type2measures["X"], dimensions
        ).execute(df)
        if len(dimensions) > 0 and len(var_measures) > 0:
            results["selection"]["distribution"] = DotPlotAggregator(
                var_measures=var_measures, dimensions=[dimensions]
            ).execute(df)
        results["selection"]["count"] = df.shape[0]

    return results


def handle_selection_ids(dataset_api, dataset, data_filter):
    df = get_selected_data(dataset_api, dataset, measures=["obs/index"], data_filter=data_filter)
    return {"ids": IdsAggregator().execute(df)}


def get_adata(
    dataset_api,
    dataset,
    embeddings=[],
    measures=[],
    dimensions=[],
    data_filters=[],
    dataset_info=None,
):
    type2measures = get_type_to_measures(measures)
    var_keys_filter = []
    obs_keys_filter = []
    selected_points_filter_basis_list = []

    for data_filter in data_filters:
        _var_keys_filter, _obs_keys_filter, selected_points_filter_basis_list = data_filter_keys(
            data_filter, dataset_info
        )

        selected_points_filter_basis_list += selected_points_filter_basis_list
        var_keys_filter += _var_keys_filter
        obs_keys_filter += _obs_keys_filter
    obs_keys = list(set(type2measures["obs"] + obs_keys_filter + dimensions))
    var_keys = list(set(type2measures["X"] + var_keys_filter))
    basis_keys = set()
    for embedding in embeddings:
        basis_keys.add(embedding["name"])
    for embedding in selected_points_filter_basis_list:
        basis_keys.add(embedding["name"])

    adata = dataset_api.read_dataset(
        dataset=dataset, keys=dict(obs=obs_keys, X=var_keys, basis=list(basis_keys))
    )
    # for basis_obj in basis_objs:
    #     if not basis_obj['precomputed'] and basis_obj['nbins'] is not None:
    #         EmbeddingAggregator.convert_coords_to_bin(df=df,
    #                                                   nbins=basis_obj['nbins'],
    #                                                   bin_name=basis_obj['full_name'],
    #                                                   coordinate_columns=basis_obj['coordinate_columns'])
    return adata


def get_selected_data(
    dataset_api,
    dataset,
    embeddings=[],
    measures=[],
    dimensions=[],
    data_filter=None,
    dataset_info=None,
):
    df = get_adata(
        dataset_api,
        dataset,
        embeddings,
        measures,
        dimensions,
        [data_filter] if data_filter is not None else [],
        dataset_info=dataset_info,
    )
    df = apply_filter(df, data_filter)
    return df


def data_filter_keys(data_filter, dataset_info=None):
    var_keys = set()
    obs_keys = set()
    basis_keys = set()
    if data_filter is not None:
        user_filters = data_filter.get("filters", [])

        for i in range(len(user_filters)):
            user_filter = user_filters[i]
            key = user_filter["field"]
            if isinstance(key, dict):
                basis_keys.add(key["name"])
            elif key == "__index":
                continue
            else:
                name, key_type = get_var_name_type(
                    key, default_type="obs", dataset_info=dataset_info
                )
                user_filter["field"] = name
                if key_type == "X":
                    var_keys.add(name)
                else:
                    obs_keys.add(name)
    return list(var_keys), list(obs_keys), list(basis_keys)



import json

import zarr

from cirrocumulus.abstract_backed_dataset import AbstractBackedDataset
from cirrocumulus.anndata_util import dataset_schema


class ZarrDataset(AbstractBackedDataset):
    def __init__(self):
        super().__init__()

    def get_suffixes(self):
        return ["zarr"]

    def is_group(self, node):
        return isinstance(node, zarr.hierarchy.Group)

    def open_group(self, filesystem, path):
        return zarr.open_group(filesystem.get_mapper(path), mode="r")

    def slice_dense_array(self, X, indices):
        return X.get_orthogonal_selection((slice(None), indices))

    def get_schema(self, filesystem, path):
        g = zarr.open_group(filesystem.get_mapper(path), mode="r")
        if "cirro-schema" in g["uns"]:
            s = str(g["uns"]["cirro-schema"][()])
            return json.loads(s)
        return dataset_schema(g, n_features=0)



import os
import json
import concurrent.futures

import numpy as np
import pandas as pd
import pyarrow as pa
import scipy.sparse
import pyarrow.parquet as pq
from anndata import AnnData

from cirrocumulus.abstract_dataset import AbstractDataset
from cirrocumulus.anndata_util import ADATA_LAYERS_UNS_KEY


max_workers = min(12, pa.cpu_count())
executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)


def read_table(path, filesystem, columns=None):
    return pq.read_table(path, filesystem=filesystem, columns=columns, use_threads=False)


def read_tables(paths, filesystem, columns=None):
    futures = []
    for path in paths:
        future = executor.submit(read_table, path, filesystem=filesystem, columns=columns)
        futures.append(future)
    concurrent.futures.wait(futures)
    return futures


def get_matrix(futures, shape=None):
    data = []
    row = []
    col = []
    is_sparse = None
    for i in range(len(futures)):
        t = futures[i].result()
        if i == 0:
            is_sparse = "index" in t.column_names
        if is_sparse:
            row.append(t.column("index").to_numpy())
            col.append(np.repeat(i, len(t)))
        data.append(t.column("value").to_numpy())
    if len(data) == 0:
        raise ValueError("No data")
    if len(row) > 0:  # sparse
        data = np.concatenate(data)
        row = np.concatenate(row)
        col = np.concatenate(col)
        # X = scipy.sparse.coo_matrix((data, (row, col)), shape=(shape[0], len(keys))).to_csc()
        # X = scipy.sparse.csr_matrix((data, (row, col)), shape=(shape[0], len(var_keys)))
        X = scipy.sparse.csc_matrix((data, (row, col)), shape=(shape[0], len(futures)))
    else:
        X = np.array(data).T
    return X


def read_matrix(keys, node_path, dataset_info, filesystem, shape):
    if len(keys) == 1 and isinstance(
        keys[0], slice
    ):  # special case if slice specified for performance
        get_item_x = keys[0]
        keys = dataset_info["var"][get_item_x]

    paths = [node_path + "/" + key + ".parquet" for key in keys]
    X = get_matrix(read_tables(paths, filesystem), shape)
    var = pd.DataFrame(index=keys)
    return X, var


class ParquetDataset(AbstractDataset):
    def __init__(self):
        super().__init__()

    def get_suffixes(self):
        return ["parquet", "pq", "cpq"]

    def read_data_sparse(self, filesystem, path, keys, dataset=None):
        dataset_info = self.get_dataset_info(filesystem, path)
        shape = dataset_info["shape"]
        X = None
        obs = None
        var = None
        obsm = {}
        X_keys = keys.pop("X", [])
        obs_keys = keys.pop("obs", [])
        basis_keys = keys.pop("basis", [])
        keys.pop("module", [])
        layers = {}
        for layer_key in keys.keys():
            X_layer, var_layer = read_matrix(
                keys=keys[layer_key],
                node_path=os.path.join(path, "layers", layer_key),
                dataset_info=dataset_info,
                filesystem=filesystem,
                shape=shape,
            )
            adata_layer = AnnData(X=X_layer, var=var_layer)
            layers[layer_key] = adata_layer

        if len(X_keys) > 0:
            X, var = read_matrix(
                keys=X_keys,
                node_path=os.path.join(path, "X"),
                dataset_info=dataset_info,
                filesystem=filesystem,
                shape=shape,
            )
        if len(obs_keys) > 0:
            obs = pd.DataFrame()
            node_path = os.path.join(path, "obs")
            paths = [node_path + "/" + key + ".parquet" for key in obs_keys]
            futures = read_tables(paths, filesystem, columns=["value"])
            for i in range(len(futures)):
                df = futures[i].result().to_pandas()
                obs[obs_keys[i]] = df["value"]

        if len(basis_keys) > 0:
            node_path = os.path.join(path, "obsm")
            paths = [node_path + "/" + key + ".parquet" for key in basis_keys]
            futures = read_tables(paths, filesystem)
            for i in range(len(futures)):
                table = futures[i].result()
                vals = []
                for c in table.column_names:
                    vals.append(table.column(c))
                vals = np.array(vals).T
                obsm[basis_keys[i]] = vals
                if X is None:
                    X = scipy.sparse.coo_matrix((vals.shape[0], 0))
        if X is None and obs is None and len(obsm.keys()) == 0:
            obs = pd.DataFrame(index=pd.RangeIndex(dataset_info["shape"][0]).astype(str))
        adata = AnnData(X=X, obs=obs, var=var, obsm=obsm)
        adata.uns[ADATA_LAYERS_UNS_KEY] = layers
        return adata

    def read_data_dense(self, filesystem, path, keys=None, dataset=None):
        var_keys = keys.pop("X", [])
        obs_keys = keys.pop("obs", [])
        basis = keys.pop("basis", [])
        all_keys = obs_keys + var_keys
        X = None
        obs = None
        var = None
        obsm = {}
        for key in keys.keys():
            all_keys += keys[key]

        if len(basis) > 0:
            for key in basis:
                # 2d only
                all_keys += ["{}_{}".format(key, 1), "{}_{}".format(key, 2)]
        df = read_table(path, filesystem=filesystem, columns=all_keys)
        if len(var_keys) > 0:
            X = df[var_keys]
        if len(obs_keys) > 0:
            obs = df[obs_keys]
        if len(basis) > 0:
            for key in basis:
                obsm[key] = df[["{}_{}".format(key, 1), "{}_{}".format(key, 2)]]
        return AnnData(X=X, obs=obs, var=var, obsm=obsm)

    def read_dataset(self, filesystem, path, keys=None, dataset=None):
        if keys is None:
            keys = {}
        # path is directory
        keys = keys.copy()
        if not path.endswith(".parquet"):
            return self.read_data_sparse(filesystem, path, keys, dataset)
        # single parquet file containing everything
        return self.read_data_dense(filesystem, path, keys, dataset)

    def get_schema(self, filesystem, path):
        if path.endswith(".json") or path.endswith(".json.gz"):
            return super().get_schema(filesystem, path)
        elif path.endswith(".parquet"):
            metadata = None
            with filesystem.open(path, "rb") as s:
                parquet_file = pq.ParquetFile(s)
                schema = parquet_file.schema.to_arrow_schema()
                result = {"version": "1"}
                for metadata_key in [b"pegasus"]:
                    if metadata_key in schema.metadata:
                        metadata = json.loads(schema.metadata[metadata_key])
                        break
                obs = []
                obs_cat = []
                if metadata is None:
                    # obs, the obsm, then var
                    names = parquet_file.schema.names
                    section = "obs"
                    result["embeddings"] = []
                    result["var"] = []
                    embedding_basename_to_count = {}
                    for name in names:
                        if section == "obs" and name.startswith("X_"):
                            section = "obsm"
                        if section == "obs":
                            field = schema.field(name)
                            if isinstance(field.type, pa.lib.DictionaryType):
                                obs_cat.append(name)
                            else:
                                obs.append(name)
                        if section == "obsm" and not name.startswith("X_"):
                            section = "var"
                        if section == "obsm":  # X_fitsne_1', 'X_fitsne_2, etc
                            basename = name[0 : name.rindex("_")]
                            count = embedding_basename_to_count.get(basename, 0)
                            embedding_basename_to_count[basename] = count + 1
                        if section == "var":
                            result["var"].append(name)
                    for name in embedding_basename_to_count.keys():
                        result["embeddings"].append(dict(name=name, dimensions=2))
                        if embedding_basename_to_count[name] == 3:
                            result["embeddings"].append(dict(name=name, dimensions=3))
                else:
                    all_obs = metadata["obs"]

                    for name in all_obs:
                        field = schema.field(name)
                        if isinstance(field.type, pa.lib.DictionaryType):
                            obs_cat.append(name)
                        else:
                            obs.append(name)
                    result["var"] = metadata["var"]
                    result["embeddings"] = metadata["obsm"]

                result["obs"] = obs
                result["obsCat"] = obs_cat
                result["shape"] = (parquet_file.metadata.num_rows, len(result["var"]))
                return result
        else:  # directory
            return super().get_schema(filesystem, os.path.join(path, "index.json.gz"))
